import re
from typing import Tuple


def split_fv(line, control_lines) -> Tuple[str, str]:
    """
    文字列をfield, valueに分割

    Args:
        line: control_linesの1行
        control_lines: controlファイルのテキスト

    Returns:
        Tuple[str, str]: field, value
    """
    field, value = line.split(": ", 1)
    # valueが複数行にまたがれる形式だった場合
    # <text></text>で包む
    if field == "Description":
        value = "<text>" + value
        for text in control_lines[control_lines.index(line) + 1 :]:
            if text[0].isspace():
                value += "\n"
                value += text
            else:
                break
        value += "</text>"
    return field, value


def make_dict(field, value, control_dict):
    """
    field, dictから辞書を生成

    Args:
        field: controlのfield
        value: controlのvalue
        control_dict: 入力する辞書
    """
    # 仮想パッケージの扱いどうしよう？showpkgでProvidesは確認できるが。。(未実装)
    # パッケージ群のvalueをリスト化・整形して辞書に入力
    if field in [
        "Depends",
        "Suggests",
        "Pre-Depends",
        "Recommends",
        "Enhances",
        "Breaks",
        "Conflicts",
        "Build-Depends",
        "Build-Depends-Indep",
        "Build-Conflicts",
        "Build-Conflicts-Indep",
    ]:
        control_dict[field] = [i for i in re.split(", ", value) if i]
    else:
        control_dict[field] = [value]


def control_to_dict(control_text: str) -> dict[str, list[str]]:
    """
    controlファイルの文字列を辞書形式に変換

    Args:
        control_text: controlファイルのテキスト

    Returns:
        dict: 変換した辞書
    """
    control_lines = control_text.splitlines()

    # 初期化
    control_dict: dict[str, list[str]] = {}
    control_dict["Depends"] = []

    # 1行ずつ分解
    for line in control_lines:
        # fieldがある行に対してのみ実行
        if line and (not line[0].isspace()) and (": " in line):
            field, value = split_fv(line, control_lines)
            make_dict(field, value, control_dict)

    # 依存関係系をまとめる
    for k in list(control_dict.keys()):
        if k in ["Suggests", "Pre-Depends", "Recommends"]:
            control_dict["Depends"] += control_dict.pop(k)

    return control_dict



def dict_to_tv(tv_dict: dict[str, list[dict[str, list[str]]]]) -> str:
    """
    SPDXのdictをTagValue形式のtextに変換

    Args:
        tv_dict: dict[str, list[dict[str, list[str]]]]: SPDXのdict

    Returns:
        str: TagValue形式のtext
    """

    tv_text = ""

    for info_name, dict_list in tv_dict.items():
        tv_text += "## " + info_name + "\n\n"

        for elem_dict in dict_list:
            for tag, value_list in elem_dict.items():
                for value in value_list:
                    tv_text += tag + ": " + value + "\n"
            else:
                tv_text += "\n\n"
    else:
        return tv_text



import argparse
import json
import os
import re
import subprocess
import sys
import time

from debiantospdx import deb_spdx
from debiantospdx.search import (
    print_package_info,
    print_spdx_files_info,
    take_spdx_path,
)


def make_pv_vrp_dict(pv_dict: dict[str, str], vrp_dict: dict[str, list[list[str]]]):
    """
    pv_dict と vrp_dict の作成

    Args:
        pv_dict: パッケージとバージョンの対応辞書
        vrp_dict: Virtual と Replace の辞書
    """
    dpkg_status_list = subprocess.run(
        ["dpkg-query", "-W", "-f=${Package}\t${Provides}\t${Replaces}\t${Version}\n"], capture_output=True, text=True
    ).stdout.splitlines()

    # Virtual と Replace の辞書作成
    for dpkg_s in dpkg_status_list:

        # dpkg_s_list: Package, Provides, Replaces, Version
        dpkg_s_list = dpkg_s.strip().split("\t")

        pv_dict[dpkg_s_list[0]] = dpkg_s_list[3]

        if dpkg_s_list[1]:
            add_vrp_dict(dpkg_s_list[1], dpkg_s_list[0], vrp_dict)
        if dpkg_s_list[2]:
            add_vrp_dict(dpkg_s_list[2], dpkg_s_list[0], vrp_dict)

    # 重複削除
    for key, value in vrp_dict.items():
        vrp_dict[key] = [list(j) for j in list(set([tuple(i) for i in value]))]


def add_vrp_dict(vrp_names: str, p_name: str, vrp_dict: dict[str, list[list[str]]]):
    """
    Virtual と Replace の辞書に情報を追加

    Args:
        vrp_names: 代替できるパッケージの名前の列記
        p_name: リアルなパッケージの名前
        vrp_dict: Virtual と Replace の辞書
    """
    vrp_name_list = vrp_names.split(", ")
    for vrp_name in vrp_name_list:
        # vrp_name_split: [Virtual or Replace Package, (比較演算子, Version)]
        vrp_name_split = [i for i in re.split(" |\\(|\\)|\\[.*?\\]", vrp_name) if i]

        if vrp_name_split[0] in vrp_dict:
            vrp_dict[vrp_name_split[0]].append([p_name] + vrp_name_split[1:])
        else:
            vrp_dict[vrp_name_split[0]] = [[p_name] + vrp_name_split[1:]]


def main(package, person, organization, all_analyze, path, search, mode, dep_mode, printinfo) -> None:
    time_start = time.perf_counter()

    pv_dict: dict[str, str] = {}  # {p_name: version}
    vrp_dict: dict[str, list[list[str]]] = {}  # {vrp_name: [[p_name, c_operator, version]]}

    make_pv_vrp_dict(pv_dict, vrp_dict)  # type: ignore

    # ディレクトリ記憶
    cwd = os.getcwd()

    os.chdir(path)

    if search is not None:
        # パッケージ情報の出力
        print_package_info(search)
    elif printinfo:
        # SPDX file群の情報出力
        print_spdx_files_info()
    elif all_analyze:
        os.mkdir("ALL")
        os.chdir("ALL")
        with open("rp_times.json", "w") as f:
            json.dump({}, f, indent=4)
        # すべてのパッケージ解析
        for p_name in pv_dict:
            if not take_spdx_path(p_name):
                deb_class = deb_spdx.DebSpdx(pv_dict, vrp_dict, p_name, person, organization, [], mode, mode)
                deb_class.init_treated_list()
                deb_class.run()
    else:
        # 特定のパッケージの解析
        if package not in pv_dict:
            print(package, "is Not exist", file=sys.stderr)
        else:
            os.mkdir(package)
            os.chdir(package)
            with open("rp_times.json", "w") as f:
                json.dump({}, f, indent=4)
            deb_class = deb_spdx.DebSpdx(pv_dict, vrp_dict, package, person, organization, [], mode, dep_mode)
            deb_class.init_treated_list()
            deb_class.run()

    os.chdir(cwd)

    time_finish = time.perf_counter()

    print("\ndebiantospdx finish")
    print("time: ", time_finish - time_start)


# 引数の処理 （エントリーポイント）
def entry():
    parser = argparse.ArgumentParser(
        prog="debiantospdx",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""Add directory path and some option.

    Analysis mode has 4 options
        0: Do not analyze files
        1: Analyze the file, but do not analyze licenses
        2: Analyze licenses of copyright file only
        3: Analyze licenses of all files""",
    )
    # 必須の位置引数
    parser.add_argument("path", type=str, help="Path of directory where SPDX files are located")

    # オプション引数 変数名は長い方になる（p, oは省略形）
    parser.add_argument("-p", "--person", nargs="+", type=str, help="The person that created the SPDX file")
    parser.add_argument("-o", "--organization", nargs="+", type=str, help="The organization that created the SPDX file")

    parser.add_argument(
        "-m",
        "--mode",
        default=2,
        type=int,
        help="Analysis mode for the specified package or all packages (default = 2)",
    )
    parser.add_argument(
        "-d",
        "--dep_mode",
        default=1,
        type=int,
        help="Analysis mode for the dependent packages (default = 1)",
    )

    # 排他的で必須なオプション引数
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("--package", type=str, help="Analyze specified package")
    group.add_argument("--all", action="store_true", help="Analyze all Debian packages")
    group.add_argument("--search", type=str, help="Print package information from SPDX files")
    group.add_argument("--printinfo", action="store_true", help="Print SPDX files' information")

    args = parser.parse_args()

    args_person = None
    args_organization = None
    if args.person is not None:
        args_person = " ".join(args.person)
    if args.organization is not None:
        args_organization = " ".join(args.organization)
    if ((args.package is not None) or args.all) and ((args.person is None) or (args.organization is None)):
        raise Exception("Enter Creator Name: person or organization")
    if not (0 <= args.mode <= 3 and 0 <= args.dep_mode <= 3):
        raise ValueError("Undefined mode")

    main(
        path=args.path,
        person=args_person,
        organization=args_organization,
        mode=args.mode,
        dep_mode=args.dep_mode,
        package=args.package,
        all_analyze=args.all,
        search=args.search,
        printinfo=args.printinfo,
    )



import hashlib
import json
import os
import re
import subprocess
import uuid

from debiantospdx import control_to_dict, dict_to_tv, make_tv_dict
from debiantospdx.search import take_spdx_path


class DebSpdx:

    pv_dict: dict[str, str]  # {p_name: version}
    vrp_dict: dict[str, list[list[str]]]  # {vrp_name: [[p_name, c_operator, version]]}
    tv_dict: dict[str, list[dict[str, list[str]]]]  # SPDX.json 参照
    control_dict: dict[str, list[str]]  # {field_name: [values]}
    package_name: str
    auth_name: str
    organization: str
    first_mode: int
    rest_mode: int
    rp_count: int
    trail_list: list[str]  # [p_name]
    treated_num: list[int] = [0]
    treated_list: list[str] = []
    doc_comment = """<text> Generated with debiantospdx and provided on an "AS IS" BASIS,
                        WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
                        No content created from debiantospdx should be considered or used as legal advice.
                        Consult an Attorney for any legal advice. </text>"""

    def return_spdx(self):
        return self.tv_dict

    def __init__(
        self, pv_dict, vrp_dict, package_name, auth_name, organization, trail_list: list[str], first_mode=2, rest_mode=1
    ):
        """
        初期化
        """
        self.pv_dict = pv_dict
        self.vrp_dict = vrp_dict
        self.tv_dict = {}
        self.control_dict = {}
        self.package_name = package_name
        self.auth_name = auth_name
        self.organization = organization
        self.trail_list = trail_list.copy()
        self.first_mode = first_mode
        self.rest_mode = rest_mode
        self.rp_count = 0

    def init_treated_list(self):
        self.treated_list.clear()

    def rm_license_dup(self, lic_dict_list):
        """
        未定義ライセンスの重複削除

        Args:
            lic_dict_list(list[str]): 未定義ライセンスのリスト
        """
        lic_rm_dup_list = []
        for lic_dict in lic_dict_list:
            for lic_rm_dup in lic_rm_dup_list:
                if lic_rm_dup == lic_dict:
                    break
            else:
                lic_rm_dup_list.append(lic_dict)
        else:
            return lic_rm_dup_list

    def merge_tv_control(self):
        """
        spdxファイルにcontrolファイルの情報を結合
        パッケージ間の依存関係情報 以外を追加・修正
        """
        tv_dict = self.tv_dict
        control_dict = self.control_dict

        # パッケージ情報の追加・修正
        package_dict = tv_dict["Package"][0]
        package_dict["PackageName"] = control_dict["Package"]
        package_dict["PackageVersion"] = control_dict["Version"]
        package_dict["SPDXID"] = ["SPDXRef-" + package_dict["PackageName"][0].replace("+", "Plus")]
        if "Homepage" in control_dict:
            package_dict["PackageHomePage"] = control_dict["Homepage"]
        package_dict["PackageComment"] = control_dict["Description"]
        package_dict["Relationship"] = []

        # クリエーション情報の追加・修正
        cre_dict = tv_dict["Creation Information"][0]
        cre_dict["Creator"].append("Tool: debiantospdx")
        if self.auth_name is not None:
            cre_dict["Creator"].append("Person: " + self.auth_name)
        if self.organization is not None:
            cre_dict["Creator"].append("Organization: " + self.organization)

        # ドキュメント情報の追加・修正
        doc_dict = tv_dict["Document Information"][0]
        doc_dict["DocumentName"] = [control_dict["Package"][0] + "_" + control_dict["Version"][0]]
        # created と packagename と versionから生成(グローバルに参照できるアドレスがある方が望ましい)
        doc_dict["DocumentNamespace"] = [
            "http://spdx.org/spdxdocs/"
            + doc_dict["DocumentName"][0]
            + "-"
            + str(uuid.uuid5(uuid.NAMESPACE_URL, (doc_dict["DocumentName"][0] + cre_dict["Created"][0])))
        ]
        doc_dict["ExternalDocumentRef"] = []
        doc_dict["Relationship"] = [doc_dict["SPDXID"][0] + " DESCRIBES " + package_dict["SPDXID"][0]]

        doc_dict["DocumentComment"] = [self.doc_comment]

        # ファイルパス と SPDXID の修正 と Relationship の追加
        for i, file_dict in enumerate(tv_dict["File"]):
            file_dict["FileName"] = [file_dict["FileName"][0].replace(("./" + package_dict["PackageName"][0]), "", 1)]
            file_dict["SPDXID"] = [package_dict["SPDXID"][0] + "-file-" + str(i)]
            file_dict["Relationship"] = [package_dict["SPDXID"][0] + " CONTAINS " + file_dict["SPDXID"][0]]

        tv_dict["Extracted License"] = self.rm_license_dup(tv_dict["Extracted License"])

    def add_relationship(self, d_list: list[str]) -> list[str]:
        """
        依存関係を処理してRelationshipフィールドを追加

        Args:
            d_list: 依存しているパッケージのリスト
        Returns:
            list[str]: 未解決な相互依存パッケージのリスト
        """
        pv_dict = self.pv_dict
        vrp_dict = self.vrp_dict
        mutual_list: list[str] = []
        package_dict = self.tv_dict["Package"][0]
        termed_d_list = []
        not_out_list = []

        dori_list = []
        for dp in d_list:
            or_list = dp.split(" | ")
            ori_list = []
            for d in or_list:
                ori_list.append([i for i in re.split(" |\\(|\\)|\\[.*?\\]|:any", d) if i])
            dori_list.append(ori_list)

        for ori_list in dori_list:
            for or_list in ori_list:
                if or_list[0] in pv_dict and self.check_version(or_list[1:], ["=", pv_dict[or_list[0]]]):
                    real_dp_name = or_list[0]
                    break

                elif or_list[0] in vrp_dict:
                    for real_p_list in vrp_dict[or_list[0]]:
                        if self.check_version(or_list[1:], real_p_list[1:]):
                            real_dp_name = real_p_list[0]
                            self.rp_count += 1
                            break
                    else:
                        continue
                    break
            else:
                continue

            # 同じ依存先を複数回指定または自分を依存先に指定しているとき
            if real_dp_name in termed_d_list or real_dp_name == self.package_name:
                continue
            else:
                termed_d_list.append(real_dp_name)

            # 既に存在しているとき
            if spdx_path := take_spdx_path(real_dp_name):
                self.add_external_ref(spdx_path, real_dp_name)
            # このパッケージの別の枝(未出力)で調査済みのとき
            elif real_dp_name in not_out_list:
                package_dict["Relationship"].append(
                    package_dict["SPDXID"][0] + " DEPENDS_ON SPDXRef-" + real_dp_name.replace("+", "Plus")
                )
            # 上の階層のパッケージと相互依存になっているとき か すでに上の階層のパッケージの別の枝で調査済みのとき
            elif real_dp_name in self.trail_list or real_dp_name in self.treated_list:
                package_dict["Relationship"].append(
                    package_dict["SPDXID"][0] + " DEPENDS_ON SPDXRef-" + real_dp_name.replace("+", "Plus")
                )
                mutual_list.append(real_dp_name)
            else:
                snap_len_treated = len(self.treated_list)
                new_spdx = DebSpdx(
                    pv_dict,
                    vrp_dict,
                    real_dp_name,
                    self.auth_name,
                    self.organization,
                    self.trail_list,
                    self.first_mode,
                    self.rest_mode,
                )
                r_mutual_list = new_spdx.run()

                # 下の階層のパッケージと相互依存になっているとき
                if r_mutual_list != []:
                    # relationship
                    package_dict["Relationship"].append(
                        package_dict["SPDXID"][0] + " DEPENDS_ON SPDXRef-" + real_dp_name.replace("+", "Plus")
                    )
                    self.merge_spdx(new_spdx.return_spdx())
                    not_out_list += self.treated_list[snap_len_treated:]
                    mutual_list += [p for p in r_mutual_list if (p != self.package_name) and (p not in not_out_list)]
                # 問題なし
                else:
                    if os.path.exists(real_dp_name + ".spdx"):
                        self.add_external_ref(real_dp_name + ".spdx", real_dp_name)
                    else:
                        self.add_external_ref(real_dp_name + ".Cycle.spdx", real_dp_name)

        self.tv_dict["Document Information"][0]["ExternalDocumentRef"] = list(
            set(self.tv_dict["Document Information"][0]["ExternalDocumentRef"])
        )
        mutual_list = list(set(mutual_list))
        return mutual_list

    def add_external_ref(self, spdx_path: str, p_name: str):
        """
        外部のSPDXファイルに依存している場合のRelationshipフィールドを追加

        Args:
            spdx_path: 外部のSPDXファイルのパス
        """
        with open(spdx_path, mode="r", encoding="utf-8") as f:
            lines_strip = [s.strip() for s in f.readlines()]
        for line in lines_strip:
            if "DocumentNamespace" in line:
                ref_space = line[19:]
                break
        else:
            return

        with open(spdx_path, mode="rb") as f:
            file_data = f.read()
        hash_sha1 = hashlib.sha1(file_data).hexdigest()
        exd_list = self.tv_dict["Document Information"][0]["ExternalDocumentRef"]
        pac_dict = self.tv_dict["Package"][0]

        doc_ref = "DocumentRef-" + spdx_path[:-5]
        exd_list.append(doc_ref + " " + ref_space + " SHA1: " + hash_sha1)
        pac_dict["Relationship"].append(
            (pac_dict["SPDXID"][0] + " DEPENDS_ON " + doc_ref + ":SPDXRef-" + p_name.replace("+", "Plus"))
        )

    def merge_spdx(self, dep_tv_dict):
        """
        2つのSPDXの情報を結合

        Args:
            dep_tv_dict: 結合するSPDXの情報
        """
        self.tv_dict["Document Information"][0]["ExternalDocumentRef"] += dep_tv_dict["Document Information"][0][
            "ExternalDocumentRef"
        ]
        self.tv_dict["Package"] += dep_tv_dict["Package"]
        self.tv_dict["File"] += dep_tv_dict["File"]
        self.tv_dict["Extracted License"] += dep_tv_dict["Extracted License"]
        self.tv_dict["Extracted License"] = self.rm_license_dup(self.tv_dict["Extracted License"])

    def compare_version(self, v1, v2, c_operator) -> bool:
        try:
            subprocess.run(["dpkg", "--compare-versions", v1, c_operator, v2], check=True)
        except subprocess.CalledProcessError:
            return False
        return True

    def check_version(self, term_list, cond_list) -> bool:
        """
        バージョン制約を満たしているかの確認

        Args:
            term_list: [p_name, c_operator, version] 制約
            cond_list: [p_name, c_operator, version] 現状

        Returns:
            bool: バージョン制約を満たしていればTrue
        """
        if term_list == []:
            return True
        elif cond_list == []:
            return False

        if term_list[0] == "=":
            match cond_list[0]:
                case "=":
                    co = "eq"
                case "<=":
                    co = "le"
                case ">=":
                    co = "ge"
                case "<<":
                    co = "lt"
                case _:
                    co = "gt"
        elif term_list[0] == "<=":
            if cond_list[0] == "=" or cond_list[0] == ">=":
                co = "ge"
            elif cond_list[0] == ">>":
                co = "gt"
            else:
                return True
        elif term_list[0] == ">=":
            if cond_list[0] == "=" or cond_list[0] == "<=":
                co = "le"
            elif cond_list[0] == "<<":
                co = "lt"
            else:
                return True
        elif term_list[0] == "<<":
            if cond_list[0] == "<=" or cond_list[0] == "<<":
                return True
            else:
                co = "gt"
        else:
            if cond_list[0] == ">=" or cond_list[0] == ">>":
                return True
            else:
                co = "lt"

        return self.compare_version(term_list[1], cond_list[1], co)

    def run(self) -> list[str]:
        """
        DebianパッケージのSPDXを推移的に生成

        Returns:
            list[str]: 未解決な相互依存パッケージのリスト
        """

        package_name = self.package_name
        self.trail_list.append(package_name)

        package_status = subprocess.run(
            ["dpkg-query", "-s", package_name], capture_output=True, text=True
        ).stdout.strip()
        self.control_dict = control_to_dict.control_to_dict(package_status)

        if len(self.treated_list) == 0:
            self.tv_dict = make_tv_dict.make_tv_dict(package_name, self.first_mode)
        else:
            self.tv_dict = make_tv_dict.make_tv_dict(package_name, self.rest_mode)

        self.treated_list.append(package_name)

        self.merge_tv_control()

        self.treated_num[0] += 1
        print("\rNumber of Analyzed Pacakges: " + str(self.treated_num[0]) + "/" + str(len(self.pv_dict)), end="")

        mutual_list = self.add_relationship(self.control_dict["Depends"])

        spdx_text = dict_to_tv.dict_to_tv(self.tv_dict)

        if mutual_list == []:
            if len(self.tv_dict["Package"]) > 1:
                with open(package_name + ".Cycle.spdx", mode="w") as f:
                    f.write(spdx_text)
            else:
                with open(package_name + ".spdx", mode="w") as f:
                    f.write(spdx_text)

        # 依存関係の参照において置換した回数を保存
        with open("rp_times.json", mode="r") as f:
            rp_times: dict[str, int] = json.load(f)
        rp_times[package_name] = self.rp_count
        with open("rp_times.json", "w") as f:
            json.dump(rp_times, f, indent=4)

        return mutual_list



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['debiantospdx']

package_data = \
{'': ['*']}

install_requires = \
['scancode-toolkit==31.2.4']

entry_points = \
{'console_scripts': ['debiantospdx = debiantospdx.cli:entry']}

setup_kwargs = {
    'name': 'debiantospdx',
    'version': '0.1.4',
    'description': 'This tool generate SPDX files from your Debian system / packages',
    'long_description': '# Debian-to-SPDX\n\n[![Apache2.0 License](https://img.shields.io/badge/License-Apatch2.0-green.svg?style=for-the-badge)](https://choosealicense.com/licenses/apache-2.0/)\n\nvscode, Github, poetryを使ったpythonの環境設定とパッケージ作成のテンプレート\n\n書き方の参考にするために適当な関数を残してあるが設定ファイルをクリーンにしているため、そのままでは動かない。\n動かすにはjoblibを入れる必要がある。\n\nまた、python-templateやtemplateは作成するリポジトリやパッケージの名前に合わせて変更すること\n\n## Usage/Examples\n\n```bash\nteplate 1 3 5 -c 22 --both\n```\n\n\n## Running Tests\n\nTo run tests, run the following command\n\n```bash\npytest -q tests\n```\n\n\n## Authors\n\n- [@tk-tanab](https://github.com/tk-tanab)\n\n\n## License\n\n- [Apatch2.0](https://choosealicense.com/licenses/apache-2.0/)\n\n',
    'author': 'tk-tanab',
    'author_email': 'tk-tanab@ist.osaka-u.ac.jp',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/tk-tanab/debiantospdx',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import hashlib
import os
import shutil
import subprocess
from datetime import datetime

from debiantospdx import tv_to_dict


def analyze_file(package_name: str, file_list: list[str], scan_copyrightfile: bool):
    """
    Debianパッケージのファイルを解析

    Args:
        package_name: 対象パッケージ
        file_list: パッケージのファイルパスのリスト
        scan_copyrightfile: copyrightファイルをscanするか

    Returns:
        list: SPDXの辞書
    """
    hash_list = []
    file_dict_list = []
    copyright_filepath = ""

    for value in file_list:
        if os.path.isfile(value):  # type: ignore
            try:
                with open(value, mode="rb") as f:
                    file_data = f.read()
                hash_sha1 = hashlib.sha1(file_data).hexdigest()  # type: ignore
                hash_list.append(hash_sha1)

                if scan_copyrightfile and value.endswith("copyright"):
                    copyright_filepath = value
                    try:
                        shutil.copyfile(value, package_name + "/copyright")
                    except (FileNotFoundError, PermissionError):
                        pass
                else:
                    file_dict = {
                        "FileName": [value],
                        "SPDXID": [],
                        "FileChecksum": ["SHA1: " + hash_sha1],
                        "LicenseConcluded": ["NOASSERTION"],
                        "LicenseInfoInFile": ["NOASSERTION"],
                        "FileCopyrightText": ["NOASSERTION"],
                    }
                    file_dict_list.append(file_dict)
            except (FileNotFoundError, PermissionError):
                pass
    else:
        hash_list.sort()
        if copyright_filepath == "":
            scan_copyrightfile = False

    tv_dict = scancode(scan_copyrightfile, package_name)
    tv_dict["Package"][0].update(
        {"PackageVerificationCode": [hashlib.sha1("".join(hash_list).encode("utf-8")).hexdigest()]}
    )
    tv_dict["File"] += file_dict_list

    if scan_copyrightfile:
        tv_dict["File"][0]["FileName"] = [copyright_filepath]
        tv_dict["File"][0]["LicenseInfoInFile"] = list(set(tv_dict["File"][0]["LicenseInfoInFile"]))

    return tv_dict


def scancode(is_scan: bool, package_name: str):
    """
    scancodeの分析結果をSPDXの辞書で返す

    Args:
        package_name: 対象パッケージ
        is_scan: scancodeするか

    Returns:
        list: SPDXの辞書
    """
    # 解析なしのテンプレート
    template_tv_dict: dict[str, list[dict[str, list[str]]]] = {
        "Document Information": [
            {
                "SPDXVersion": ["SPDX-2.2"],
                "DataLicense": ["CC0-1.0"],
                "DocumentNamespace": [],
                "DocumentName": [],
                "SPDXID": ["SPDXRef-DOCUMENT"],
                "DocumentComment": [],
                "ExternalDocumentRef": [],
                "Relationship": [],
            }
        ],
        "Creation Information": [{"Creator": [], "Created": [datetime.now().strftime("%Y-%m-%dT%H:%M:%SZ")]}],
        "Package": [
            {
                "PackageName": [],
                "PackageVersion": [],
                "SPDXID": [],
                "PackageHomePage": [],
                "PackageDownloadLocation": ["NOASSERTION"],
                "PackageVerificationCode": [],
                "PackageLicenseDeclared": ["NOASSERTION"],
                "PackageLicenseConcluded": ["NOASSERTION"],
                "PackageLicenseInfoFromFiles": ["NOASSERTION"],
                "PackageCopyrightText": ["NOASSERTION"],
            }
        ],
        "File": [],
        "Extracted License": [],
    }
    if is_scan:
        output = package_name + "/output.tag"
        subprocess.run(
            ["scancode", "-clpi", package_name, "--spdx-tv", output], stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )

        tv_dict = tv_to_dict.tv_to_dict(output)
        tv_dict["Creation Information"][0]["Created"] = [datetime.now().strftime("%Y-%m-%dT%H:%M:%SZ")]
        return tv_dict
    else:
        return template_tv_dict


def make_tv_dict(package_name, mode):
    """
    SPDXの辞書を生成する

    Args:
        package_name: 対象パッケージ
        mode:
            0: ファイル解析しない
            1: ファイル解析はするが、ライセンスは解析しない
            2: copyrightファイルだけライセンス解析する
            3: 全てのファイルのライセンスを解析する

    Returns:
        list: SPDXの辞書
    """

    # 0: ファイル解析しない
    if mode == 0:
        tv_dict = scancode(False, package_name)
        tv_dict["Package"][0]["FilesAnalyzed"] = ["false"]
        tv_dict["Package"][0].pop("PackageLicenseInfoFromFiles")
        return tv_dict

    # ファイルを展開する作業用ディレクトリの作成
    os.mkdir(package_name)

    # パッケージを構成するファイルの一覧を取得
    file_list = subprocess.run(["dpkg", "-L", package_name], capture_output=True, text=True).stdout.splitlines()

    tv_dict = {}

    # 1: ファイル解析はするが、ライセンスは解析しない
    if mode == 1 or file_list == []:
        tv_dict = analyze_file(package_name, file_list, False)

    # 2: copyrightファイルだけライセンス解析する
    elif mode == 2:
        tv_dict = analyze_file(package_name, file_list, True)

    # 3: 全てのファイルのライセンスを解析する
    elif mode == 3:
        # ファイルを作業用ディレクトリに展開
        for value in file_list:
            if os.path.isdir(value):
                dirname = package_name + value
                os.makedirs(dirname, exist_ok=True)
            else:
                try:
                    shutil.copyfile(value, package_name + value)
                except (FileNotFoundError, PermissionError):
                    pass

        tv_dict = scancode(True, package_name)

    # 作業用ディレクトリの削除
    shutil.rmtree(package_name)

    return tv_dict



import glob
import json
import os


def take_spdx_path(p_name) -> str:
    """
    指定されたパッケージ情報を含むSPDXファイルの検出
    ない場合は空、ある場合はSPDXファイルのパスを返す

    Args:
        p_name: 存在するか確認するパッケージ
        version: パッケージのバージョン
    Returns:
        空 or SPDXファイルのパス
    """
    if os.path.exists(p_name + ".spdx"):
        return p_name + ".spdx"
    else:
        for spdx_path in glob.glob("*.Cycle.spdx"):
            with open(spdx_path, mode="r", encoding="utf-8") as f:
                text = f.read()
            if ("\nPackageName: " + p_name + "\nPackageVersion: ") in text:
                return spdx_path
        else:
            return ""


def pick_version(path, p_name):
    in_package_info = False
    version = ""
    with open(path, mode="r") as f:
        for line in f:
            if in_package_info:
                version = line[16:].strip()
                break
            if line == "## File\n":
                break
            if line == ("PackageName: " + p_name + "\n"):
                in_package_info = True
    return version


def take_pakages(path):
    packages = []
    with open(path, mode="r") as f:
        for line in f:
            if line.startswith("PackageName: "):
                packages.append(line[13:].strip())
            if line == "## File\n":
                break
    return packages


def take_pakage_version(path):
    packages = []
    in_package_info = False
    p_and_v = []
    with open(path, mode="r") as f:
        for line in f:
            if in_package_info:
                p_and_v.append(line[16:].strip())
                packages.append(p_and_v.copy())
                p_and_v.clear()
            if line.startswith("PackageName: "):
                p_and_v.append(line[13:].strip())
                in_package_info = True
            if line == "## File\n":
                break
    return packages


def take_expaths(path):
    expaths = []
    with open(path, mode="r") as f:
        for line in f:
            if line.startswith("ExternalDocumentRef: "):
                expaths.append(line.split()[1][12:].strip() + ".spdx")
            if line == "## Package\n":
                break
    return expaths


def make_spdx_dict() -> dict[str, list[str]]:
    spdx_dict: dict[str, list[str]] = {}
    for spdx_path in glob.glob("*.spdx"):
        spdx_dict[spdx_path] = take_expaths(spdx_path)
    else:
        return spdx_dict


def make_depend_recursive_dict(spdx_dict: dict[str, list[str]]):
    dr_dict: dict[str, list[str]] = {}
    for spdx in spdx_dict:
        if spdx not in dr_dict:
            make_dr_dict_sub(spdx, dr_dict, spdx_dict)
    else:
        return dr_dict


def make_dr_dict_sub(spdx_path: str, dr_dict: dict[str, list[str]], spdx_dict: dict[str, list[str]]):
    dr_dict[spdx_path] = spdx_dict[spdx_path].copy()
    for spdx in spdx_dict[spdx_path]:
        if spdx not in dr_dict:
            make_dr_dict_sub(spdx, dr_dict, spdx_dict)
        dr_dict[spdx_path] += dr_dict[spdx]
    else:
        dr_dict[spdx_path] = list(set(dr_dict[spdx_path]))


def take_depends_recursive(p_name: str, spdx_path: str, dr_dict: dict[str, list[str]]):
    depends = take_pakages(spdx_path)
    depends.remove(p_name)
    for ex_spdx in dr_dict[spdx_path]:
        depends += take_pakages(ex_spdx)
    else:
        return depends


def take_rdepends_recursive(p_name: str, spdx_path: str, dr_dict: dict[str, list[str]]):
    rdepends = take_pakages(spdx_path)
    rdepends.remove(p_name)
    for spdx, ex_spdxs in dr_dict.items():
        if spdx_path in ex_spdxs:
            rdepends += take_pakages(spdx)
    else:
        return rdepends


def count_cycle_depend(dr_dict: dict[str, list[str]]):
    counter = 0
    for spdx, ex_spdxs in dr_dict.items():
        times = 1
        selfcycle = 0
        if spdx.endswith("Cycle.spdx"):
            times = len(take_pakages(spdx))
            selfcycle = 1
        counter += (sum(ex_spdx.endswith("Cycle.spdx") for ex_spdx in ex_spdxs) + selfcycle) * times
    else:
        return counter


def count_cycle_spdx():
    return len(glob.glob("*.Cycle.spdx"))


def get_spdxs_size():
    sum_bytesize = 0
    for spdx in glob.glob("*.spdx"):
        sum_bytesize += os.path.getsize(spdx)
    else:
        return sum_bytesize


def count_spdx():
    return len(glob.glob("*.spdx"))


def count_packages():
    counter = 0
    for spdx in glob.glob("*.spdx"):
        counter += len(take_pakages(spdx))
    else:
        return counter


def take_files(path):
    packages = []
    with open(path, mode="r") as f:
        for line in f:
            if line.startswith("PackageName: "):
                packages.append(line[13:].strip())
            if line == "## File\n":
                break
    return packages


def count_files():
    counter = 0
    for spdx in glob.glob("*.spdx"):
        with open(spdx, mode="r") as f:
            text = f.read()
        counter += text.count("\nFileName: ")
    else:
        return counter


def count_replace(dr_dict: dict[str, list[str]]):
    with open("rp_times.json", mode="r") as f:
        rp_dict: dict[str, int] = json.load(f)

    file_rp_dict: dict[str, int] = {}
    numpackage_dict: dict[str, int] = {}
    for spdx in dr_dict:
        file_rp_dict[spdx] = 0
        packages = take_pakages(spdx)
        numpackage_dict[spdx] = len(packages)
        for package in packages:
            file_rp_dict[spdx] += rp_dict[package]

    sum_counter = 0

    for spdx, ex_spdxs in dr_dict.items():
        counter = file_rp_dict[spdx]
        for ex_spdx in ex_spdxs:
            counter += file_rp_dict[ex_spdx]
        else:
            sum_counter = counter * numpackage_dict[spdx]

    return sum_counter


def print_pvlist(pv_list):
    for pv in pv_list:
        print(" ", pv[0], "(" + pv[1] + ")")


def print_package_info(p_name):
    spdx_path = take_spdx_path(p_name)
    version = pick_version(spdx_path, p_name)
    spdx_dict = make_spdx_dict()
    dr_dict = make_depend_recursive_dict(spdx_dict)
    depends = take_depends_recursive(p_name, spdx_path, dr_dict)
    rdepends = take_rdepends_recursive(p_name, spdx_path, dr_dict)
    print("Package Name".ljust(17) + ":", p_name)
    print("Package Version".ljust(17) + ":", version)
    print("SPDX File Name".ljust(17) + ":", spdx_path, "\n")
    print("--Dependency Recursive--")
    print("---------------------------------------------")
    print_pvlist(depends)
    print("\n--Reverse Dependency Recursive--")
    print("---------------------------------------------")
    print_pvlist(rdepends)


def print_spdx_files_info():
    spdx_dict = make_spdx_dict()
    dr_dict = make_depend_recursive_dict(spdx_dict)
    print("Number of Cycle.spdx files".ljust(50) + ":", count_cycle_spdx())
    print("Number of Circulation Dependent Occurrences".ljust(50) + ":", count_cycle_depend(dr_dict))
    print("Total Bytes".ljust(50) + ":", get_spdxs_size())
    print("Number of SPDX files".ljust(50) + ":", count_spdx())
    print("Number of Packages".ljust(50) + ":", count_packages())
    print("Number of Files".ljust(50) + ":", count_files())
    print("Number of times Replace or Provide was referenced".ljust(50) + ":", count_replace(dr_dict))



from typing import Tuple


def split_tv(line: str, line_iter) -> Tuple[str, str]:
    """
    文字列をfield, valueに分割

    Args:
        line: spdx_tvの1行
        line_iter: spdx_tvの全行をiterにしたもの

    Returns:
        Tuple[str, str]: field, value
    """

    tag, value = line.split(": ", 1)
    # valueが<text>複数行にまたがれる形式だった場合
    # iterを読み進める
    if value.strip().startswith("<text>"):
        for text in line_iter:
            value += "\n"
            value += text
            if "</text>" in text:
                break
    return tag, value


# TagValue形式のspdxをjson形式に変換
def tv_to_dict(spdx_tv: str) -> dict[str, list[dict[str, list[str]]]]:
    """
    spdx_tvファイルの文字列を辞書形式に変換

    Args:
        spdx_tv: spdx_tvファイルのパス

    Returns:
        dict: 変換した辞書
    """
    # scancode-toolkitが生成したspdxを読む
    # iterにするのはtextとfieldの見分けがつかないため、
    # 条件でtextを読み飛ばすことができず、iterを送るしかないから
    with open(spdx_tv, mode="r", encoding="utf-8") as f:
        lines_strip = [s.strip() for s in f.read().strip().splitlines()]
        lines_iter = iter(lines_strip)

    spdx_dict: dict[str, list[dict[str, list[str]]]] = {}
    new_elem_dict: dict[str, list[str]] = {}
    info_name: str = ""

    for line in lines_iter:
        # infomationは新たな要素の書き出し
        if line.startswith("#"):
            info_name = line[1:].strip()

            if info_name == "Creation Info":
                info_name = "Creation Information"
            elif info_name == "Extracted Licenses":
                info_name = "Extracted License"

            if info_name not in spdx_dict:
                new_elem_dict = {}
                spdx_dict[info_name] = [new_elem_dict]
        # tag, valueの行の扱い(空行を除く)
        elif line.strip():
            tag, value = split_tv(line, lines_iter)
            if tag in new_elem_dict:
                new_elem_dict[tag].append(value)
            else:
                new_elem_dict[tag] = [value]
                if info_name == "Package":
                    if tag == "PackageName":
                        new_elem_dict["PackageVersion"] = []
                    elif tag == "SPDXID":
                        new_elem_dict["PackageHomePage"] = []
        # 空行の扱い
        elif new_elem_dict:
            new_elem_dict = {}
            spdx_dict[info_name].append(new_elem_dict)

    else:
        for info_dict_list in spdx_dict.values():
            if (last_dict := info_dict_list.pop()) != {}:
                info_dict_list.append(last_dict)

        return spdx_dict
